from __future__ import annotations

from asyncio import sleep
from asyncio.locks import Event
from asyncio.tasks import create_task, gather
from datetime import timedelta
from typing import Callable, Union

from crontab import CronTab

from donald.tasks import TaskWrapper

from .utils import SchedulerNotReadyError, logger

TInterval = Union[timedelta, int, float, str, CronTab]


class Scheduler:
    """Manage schedules."""

    def __init__(self):
        self._schedule = []
        self._tasks = []
        self._finished = None

    def start(self):
        logger.info("Starting task scheduler")
        self._finished = Event()
        for scheduler in self._schedule:
            self._tasks.append(create_task(scheduler()))

    async def stop(self):
        for task in self._tasks:
            task.cancel()

        await self.join()
        if self._finished is not None:
            self._finished.set()

    def wait(self):
        if self._finished is None:
            raise SchedulerNotReadyError
        return self._finished.wait()

    async def join(self):
        await gather(*self._tasks, return_exceptions=True)
        self._tasks.clear()

    def schedule(self, interval: TInterval) -> Callable[[TaskWrapper], TaskWrapper]:
        """Schedule a task to run periodically."""

        assert not callable(interval), "Use @manager.schedule(interval)"

        if isinstance(interval, str):
            interval = CronTab(interval)

        if isinstance(interval, timedelta):
            timer = interval.total_seconds

        elif isinstance(interval, CronTab):
            timer = lambda: interval.next(default_utc=True)  # type: ignore[union-attr]

        else:
            timer = lambda: float(interval)  # type: ignore[arg-type]

        def wrapper(task: TaskWrapper):
            assert isinstance(task, TaskWrapper), "Only tasks can be scheduled."

            async def scheduler():
                while True:
                    to_sleep = max(timer(), 1e-2)
                    logger.info("Next '%s' in %0.2f s", task.import_path(), to_sleep)
                    await sleep(to_sleep)
                    task.submit()

            self._schedule.append(scheduler)
            return task

        return wrapper



import pytest

from .tasks import async_task, manager


@pytest.fixture()
async def setup(redis_url):
    manager.setup(backend="redis", backend_params={"url": redis_url})
    assert manager._backend
    assert manager._backend.params

    async with manager:
        w1 = manager.create_worker()
        # w2 = manager.create_worker()
        w1.start()
        # w2.start()
        yield manager
        await w1.stop()
        # await w2.stop()


async def test_submit_task(setup, sleep, checklog):
    res = async_task.submit()
    assert res
    assert await res

    await sleep(1e-1)

    assert checklog("Run async_task 42")

    for n in range(3):
        assert await async_task.submit(n)

    await sleep(1e-1)
    assert checklog("Run async_task 0")
    assert checklog("Run async_task 1")
    assert checklog("Run async_task 2")



from __future__ import annotations

import os
from asyncio import iscoroutine
from asyncio.exceptions import CancelledError
from asyncio.locks import Event, Semaphore
from asyncio.tasks import Task, create_task, gather, sleep
from contextlib import suppress
from importlib import metadata
from typing import TYPE_CHECKING, AsyncIterator, Callable, ClassVar, Iterable, cast

try:
    from asyncio import timeout as async_timeout  # type: ignore[attr-defined]
except ImportError:  # python 39, 310
    from async_timeout import timeout as async_timeout  # type: ignore[assignment, no-redef]

from donald.tasks import TaskWrapper

from .types import TRunArgs, TTaskParams, TWorkerParams
from .utils import import_obj, logger

if TYPE_CHECKING:
    from numbers import Number

    from .backend import BaseBackend

BANNER = r"""

$$$$$$$\                                $$\       $$\
$$  __$$\                               $$ |      $$ |
$$ |  $$ | $$$$$$\  $$$$$$$\   $$$$$$\  $$ | $$$$$$$ |
$$ |  $$ |$$  __$$\ $$  __$$\  \____$$\ $$ |$$  __$$ |
$$ |  $$ |$$ /  $$ |$$ |  $$ | $$$$$$$ |$$ |$$ /  $$ |
$$ |  $$ |$$ |  $$ |$$ |  $$ |$$  __$$ |$$ |$$ |  $$ |
$$$$$$$  |\$$$$$$  |$$ |  $$ |\$$$$$$$ |$$ |\$$$$$$$ |
\_______/  \______/ \__|  \__| \_______|\__| \_______|
"""


class Worker:
    defaults: ClassVar[TWorkerParams] = {
        "max_tasks": 0,
        "task_defaults": None,
        "on_start": None,
        "on_stop": None,
        "on_error": None,
        "show_banner": False,
    }

    def __init__(self, backend: BaseBackend, params: TWorkerParams):
        self._backend = backend
        self._runner = None
        self._params = cast(TWorkerParams, dict(self.defaults, **params))
        self._task_params = cast(TTaskParams, self._params["task_defaults"] or {})

        self.on_error = self._params.get("on_error")

        max_tasks = self._params["max_tasks"]
        self._sem = max_tasks and Semaphore(max_tasks - 1) or None

        self._tasks: set[Task] = set()
        self._finished = Event()

    def start(self):
        """Start the worker."""
        msg = self._params.get("show_banner") and BANNER or ""
        msg += f"\n\nDonald v{metadata.version('donald')} - Worker"
        msg += f"\nBackend: {self._backend.backend_type}"
        msg += f"\nPID: {os.getpid()}\n"

        logger.info(msg)
        self._finished.clear()
        self._runner = create_task(self.run_worker(), name="Worker Run Task")
        self._runner.add_done_callback(self.finish_task)

    async def stop(self):
        """Stop the worker."""
        logger.info("Stopping worker")
        for task in self._tasks:
            task.cancel()

        if self._runner and not self._runner.done():
            self._runner.add_done_callback(lambda _: self._finished.set())
            self._runner.cancel()

        await self.join()

        on_stop = self._params.get("on_stop")
        if on_stop:
            await on_stop()

        self._finished.set()

    def wait(self):
        return self._finished.wait()

    async def run_worker(self: Worker):
        on_start = self._params.get("on_start")
        if on_start:
            await on_start()

        logger.info("Worker started")
        task_iter: AsyncIterator[TRunArgs] = await self._backend.subscribe()
        sem, tasks, finish_task = self._sem, self._tasks, self.finish_task
        async for task_msg in task_iter:
            path, args, kwargs, params = task_msg
            try:
                tw = import_obj(path)
            except Exception as exc:  # noqa: BLE001
                logger.exception("Failed to get task: %s", path, exc_info=exc)
                continue

            if not isinstance(tw, TaskWrapper):
                logger.error("Invalid task: %s", path)
                continue

            task_params = cast(TTaskParams, dict(self._task_params, **params))
            name = tw.import_path()
            task = create_task(
                self.run_task(tw, args, kwargs, **task_params),
                name=name,
            )
            tasks.add(task)
            task.add_done_callback(finish_task)
            logger.info("Run: '%s' (%d)", name, id(task))
            if sem:
                await sem.acquire()

    async def run_task(
        self,
        corofunc: Callable,
        args: Iterable,
        kwargs: dict,
        *,
        timeout: Number | None = None,
        delay: Number | None = None,
        **_,
    ):
        # Process delay
        if delay:
            await sleep(cast(float, delay))

        # Process timeout
        if timeout:
            async with async_timeout(cast(float, timeout)):
                return await corofunc(*args, **kwargs)

        return await corofunc(*args, **kwargs)

    def finish_task(self, task: Task):
        self._tasks.discard(task)

        with suppress(CancelledError):
            exc = task.exception()
            if exc:
                logger.exception(
                    "Fail: '%s' (%d)",
                    task.get_name(),
                    id(task),
                    exc_info=exc,
                )
                if self.on_error:
                    coro = self.on_error(exc)
                    if iscoroutine(coro):
                        create_task(coro)

        if self._sem:
            self._sem.release()

        logger.info("Finished: '%s' (%d)", task.get_name(), id(task))

    async def join(self):
        """Wait for all tasks to complete."""
        tasks = self._tasks
        if tasks:
            logger.info("Waiting for %d tasks to complete", len(self._tasks))
            return await gather(*tasks, return_exceptions=True)
        return None



from donald import logger

from .tasks import async_task, manager


async def test_on_start(sleep, checklog):
    @manager.on_start
    async def on_start():
        logger.info("Run on_start")

    async with manager:
        w = manager.create_worker()
        w.start()

        await async_task.submit()
        await sleep(1e-1)
        assert checklog("Run async_task 42")

        await w.stop()

    assert checklog("Run on_start")


async def test_on_stop(checklog, sleep):
    @manager.on_stop
    async def on_stop():
        logger.info("Run on_stop")

    async with manager:
        w = manager.create_worker()
        w.start()

        await async_task.submit()
        await sleep(1e-1)
        assert checklog("Run async_task 42")

        await w.stop()

    assert checklog("Run on_stop")


async def test_on_error(checklog, sleep):
    @manager.on_error
    async def on_error(exc):
        logger.info("Run on_error: %s", exc)

    async with manager:
        w = manager.create_worker()
        w.start()

        await async_task.submit(error=True)
        await sleep(1e-1)

        await w.stop()

    checklog("Run on_error: Task failed")
